@@ -1,32 +1,54 @@
 #! /usr/bin/env python
 
 """\
-A simple but flexible parser of SUSY Les Houches Accord (SLHA) model and decay files.
+A simple but flexible handler of the SUSY Les Houches Accord (SLHA) data format.
 
 pyslha is a parser/writer module for particle physics SUSY Les Houches Accord
 (SLHA) supersymmetric spectrum/decay files, and a collection of scripts which
 use the interface, e.g. for conversion to and from the legacy ISAWIG format, or
 to plot the mass spectrum and decay chains.
 
-The current release supports SLHA version 1, and as far as we're aware is also
+The current release supports SLHA version 1, and as far as I'm aware is also
 fully compatible with SLHA2: the block structures are read and accessed
-completely generically. If you have any problems with SLHA2, please provide an
-example input file and we'll investigate.
+generically. If you have any problems, please provide an example input file and
+I'll happily investigate.
 
 The plotting script provides output in PDF, EPS and PNG via LaTeX and the TikZ
 graphics package, and as LaTeX/TikZ source for direct embedding into documents or
 user-tweaking of the generated output.
 
+Users of version 1.x should note that the interface has changed a little in
+version 2.0.0 and onward, in particular in the interface of the Block objects,
+which are now more dict-like: entries can be added and accessed via the usual
+square-brackets indexing operators, including for multiple indices as is common
+for mixing matrices e.g. NMIX[1,2] as opposed to the old NMIX.entries[1][2]
+way. This does break backward compatibility but is a big improvement both for
+internal code sanity and usability. The Block interface also now supplies
+dict-like has_key(), keys(), and items() methods, as well as more specialist
+value(), set_value() and is_single_valued() methods for improved access to ALPHA
+and any other unindexed blocks.
+
+Please cite PySLHA via the URL http://insectnation.org/projects/pyslha if it is
+useful to you in the production of an academic paper. A PySLHA publication will
+be provided at some point... watch this space!
+
+
 TODOs:
- * Store dict entries as tuple keys, e.g. myblock.entries[1,2] rather than recursive dicts => v1.6.0
- * Preserve comments from read -> write (needs full-line/inline comment separation?)
- * Split writeSLHA into writeSLHA{Blocks,Decays}
- * Identify HERWIG decay matrix element to use in ISAWIG
- * Handle RPV SUSY in ISAWIG
+
+  For 2.1.0:
+   * Preserve comments from read -> write (needs full-line/inline comment
+     separation?). Can use separate comment dicts in Block and Decay, and
+     attach a multiline .comment attr to the returned/written dicts.
+   * Only allow integers to be used as block entry indices, to allow multiple
+     values (request, if possible, from Anders & Are).
+
+  Later/maybe:
+   * Identify HERWIG decay matrix element to use in ISAWIG
+   * Handle RPV SUSY in ISAWIG
 """
 
 __author__ = "Andy Buckley <andy.buckley@cern.ch"
-__version__ = "1.5.0"
+__version__ = "2.0.0"
 
 
 def _mkdict():
@@ -54,12 +76,19 @@ def _autotype(var):
         return var
 
 def _autostr(var, precision=8):
-    """Automatically numerical types to the right sort of string."""
+    """Automatically format numerical types as the right sort of string."""
     if type(var) is float:
-        return ("%." + str(precision) + "e") % var
+        return ("% ." + str(precision) + "e") % var
     return str(var)
 
 
+class AccessError(Exception):
+    "Exception object to be raised when a SLHA block is accessed in an invalid way"
+    def __init__(self, errmsg):
+        self.msg = errmsg
+    def __str__(self):
+        return self.msg
+
 class ParseError(Exception):
     "Exception object to be raised when a spectrum file/string is malformed"
     def __init__(self, errmsg):
@@ -71,47 +100,139 @@ class ParseError(Exception):
 
 class Block(object):
     """
-    Object representation of any BLOCK elements read from the SLHA file.  Blocks
-    have a name, may have an associated Q value, and then a collection of data
-    entries, stored as a recursive dictionary. Types in the dictionary are
-    numeric (int or float) when a cast from the string in the file has been
-    possible.
+    Object representation of any BLOCK elements read from an SLHA file.
+
+    Blocks have a name, may have an associated Q value, and contain a collection
+    of data entries, each indexed by one or more keys. Entries in the dictionary
+    are stored as numeric types (int or float) when a cast from the string in
+    the file has been possible.
+
+    Block is closely related to a Python dict (and, in fact, is implemented via
+    an OrderedDict when possible). The preferred methods of entry access use the
+    dict-like [] operator for getting and setting, and the keys() and items()
+    methods for iteration. Purely iterating over the object behaves like keys(),
+    as for an ordinary dict.
+
+    Multiple indices are possible, especially for entries in mixing matrix
+    blocks. These are now implemented in the natural way, e.g. for access to the
+    (1,2) element of a mxing matrix block, use bmix[1,2] = 0.123 and print
+    bmix[1,2]. The value() and set_value() functions behave similarly.
+
+    It is possible, although not usual, to store unindexed values in a
+    block. This is only supported when that entry is the only one in the block,
+    and it is stored in the normal fashion but with None as the lookup key. The
+    value() method may be used without a key argument to retrieve this value, if
+    the is_single_valued() method returns True, and similarly the set_value()
+    method may be used to set it if only one argument is supplied and the object
+    is compatible.
     """
-    def __init__(self, name, q=None):
+    def __init__(self, name, q=None, entries=None):
         self.name = name
         self.entries = _mkdict()
+        if entries is not None:
+            self.entries.update(entries)
         self.q = _autotype(q)
 
     def add_entry(self, entry):
-        #print entry
-        nextparent = self.entries
+        """Add an entry to the block from an iterable (i.e. list or tuple) or
+        whitespace-separated string.
+
+        Indexing will be determined automatically such that there is always a
+        single-element value: multi-value or None indices may be constructed
+        implicitly.
+        """
+        ## If the argument is a single string, split it and proceed
         if type(entry) is str:
-            raise Exception("Block entries must be tuples or lists")
-        if len(entry) < 2:
-            raise Exception("Block entry tuples must have at least two entries")
-        #print "in", entry
+            entry = entry.split()
+        ## Check that the arg is an iterable
+        if not hasattr(entry, "__iter__"):
+            raise AccessError("Block entries must be iterable")
+        ## Auto-convert the types in the list
         entry = map(_autotype, entry)
-        #print "out", entry
-        for e in entry[:-2]:
-            if e is not entry[-1]:
-                nextparent = nextparent.setdefault(e, _mkdict())
-        nextparent[entry[-2]] = entry[-1]
-        #print self.entries
+        ## Add the entry to the map, with appropriate indices
+        if len(entry) == 1:
+            self.entries[None] = entry[0]
+        elif len(entry) == 2:
+            self.entries[entry[0]] = entry[1]
+        else:
+            self.entries[tuple(entry[:-1])] = entry[-1]
+
+    def is_single_valued(self):
+        """Return true if there is only one entry, and it has no index: the
+        'value()' attribute may be used in that case without an argument."""
+        return len(self.entries) == 1 and self.entries.keys()[0] is None
+
+    def value(self, key=None):
+        """Get a value from the block with the supplied key.
+
+        If no key is given, then the block must contain only one non-indexed
+        value otherwise an AccessError exception will be raised.\
+        """
+        if key == None and not self.is_single_valued():
+            raise AccessError("Tried to access unique value of multi-value block")
+        return self.entries[key]
+
+    def set_value(self, *args):
+        """Set a value in the block with supplied key, val args (or just val for
+        an unindexed block).
+
+        If no key is given, then the block must contain only one non-indexed
+        value otherwise an AccessError exception will be raised.\
+        """
+        if len(args) == 0:
+            raise AccessError("set_value() called without arguments")
+        elif len(args) == 1:
+            if len(self.entries) > 0 and not self.is_single_valued():
+                raise AccessError("Tried to set a unique value on a multi-value block")
+            self.entries[None] = args[0]
+        else:
+            self.entries[tuple(args[:-1])] = args[-1]
+
+    def has_key(self, key):
+        """Does the block have the given key?"""
+        return self.entries.has_key(key)
+
+    def keys(self):
+        """Access the block item keys."""
+        return self.entries.keys()
+
+    def items(self, key=None):
+        """Access the block items as (key,value) tuples.
+
+        Note: The Python 3 dict attribute 'items()' is used rather than the
+        'old' Python 2 'iteritems()' name for forward-looking compatibility.\
+        """
+        return self.entries.iteritems()
+
+    def __len__(self):
+        return len(self.entries)
+
+    def __iter(self):
+        return self.entries.__iter__()
+
+    def __getitem__(self, key):
+        return self.entries[key]
+
+    def __setitem__(self, key, value):
+        self.entries[key] = value
 
     def __cmp__(self, other):
-        return cmp(self.name, other.name)
+        return cmp(self.name, other.name) and cmp(self.entries, other.entries)
 
-    def __str__(self):
+    def __repr__(self):
         s = self.name
         if self.q is not None:
             s += " (Q=%s)" % self.q
-        s += "\n"
-        s += str(self.entries)
+        def _format_kv(k, v):
+            if type(k) is not tuple:
+                s = "%r" % k
+            else:
+                s = ",".join("%r" % subindex for subindex in k)
+            s += " : %r" % v
+            return s
+        s += " { " + "; ".join(_format_kv(k, v) for k, v in self.items()) + " }"
         return s
 
-    def __repr__(self):
-        return self.__str__()
-
 
 class Decay(object):
     """
@@ -135,11 +256,8 @@ class Decay(object):
     def __cmp__(self, other):
         return cmp(other.br, self.br)
 
-    def __str__(self):
-        return "%.8e %s" % (self.br, self.ids)
-
     def __repr__(self):
-        return self.__str__()
+        return "%.8e %s" % (self.br, self.ids)
 
 
 class Particle(object):
@@ -169,7 +287,7 @@ class Particle(object):
             return cmp(self.pid, other.pid)
         return cmp(abs(self.pid), abs(other.pid))
 
-    def __str__(self):
+    def __repr__(self):
         s = str(self.pid)
         if self.mass is not None:
             s += " : mass = %.8e GeV" % self.mass
@@ -180,12 +298,6 @@ class Particle(object):
                 s += "\n  %s" % d
         return s
 
-    def __repr__(self):
-        return self.__str__()
-
-
-
-
 
 def readSLHA(spcstr, ignorenobr=False):
     """
@@ -203,6 +315,7 @@ def readSLHA(spcstr, ignorenobr=False):
     currentdecay = None
     for line in spcstr.splitlines():
         ## Handle (ignore) comment lines
+        # TODO: Store block/entry comments
         if line.startswith("#"):
             continue
         if "#" in line:
@@ -237,12 +350,7 @@ def readSLHA(spcstr, ignorenobr=False):
                 if len(items) < 1:
                     continue
                 if currentblock != "DECAY":
-                    if len(items) < 2:
-                        ## Treat the ALPHA block differently
-                        blocks[currentblock].value = _autotype(items[0])
-                        blocks[currentblock].entries = _autotype(items[0])
-                    else:
-                        blocks[currentblock].add_entry(items)
+                    blocks[currentblock].add_entry(items)
                 else:
                     br = float(items[0]) if items[0].upper() != "NAN" else None
                     nda = int(items[1])
@@ -253,9 +361,9 @@ def readSLHA(spcstr, ignorenobr=False):
     ## Try to populate Particle masses from the MASS block
     # print blocks.keys()
     try:
-        for pid in blocks["MASS"].entries.keys():
+        for pid in blocks["MASS"].keys():
             if decays.has_key(pid):
-                decays[pid].mass = blocks["MASS"].entries[pid]
+                decays[pid].mass = blocks["MASS"][pid]
     except:
         raise ParseError("No MASS block found: cannot populate particle masses")
 
@@ -263,43 +371,52 @@ def readSLHA(spcstr, ignorenobr=False):
 
 
 
-
-# TODO: Split writeSLHA into writeSLHA{Blocks,Decays}
-
-
-def writeSLHA(blocks, decays, ignorenobr=False, precision=8):
-    """
-    Return an SLHA definition as a string, from the supplied blocks and decays dicts.
-    """
-    fmte = "%." + str(precision) + "e"
-
-    sep = "   "
-    out = ""
-    def dict_hier_strs(d, s=""):
-        if isinstance(d, dict):
-            for k, v in d.iteritems():
-                for s2 in dict_hier_strs(v, s + sep + _autostr(k)):
-                    yield s2
-        else:
-            yield s + sep + _autostr(d)
-    ## Blocks
+def writeSLHABlocks(blocks, precision=8):
+    """Return an SLHA definition as a string, from the supplied blocks dict."""
+    sep = 3 * " "
+    blockstrs = []
     for bname, b in blocks.iteritems():
         namestr = b.name
         if b.q is not None:
-            namestr += (" Q= " + fmte) % float(b.q)
-        out += "BLOCK %s\n" % namestr
-        for s in dict_hier_strs(b.entries):
-            out += sep + s + "\n"
-        out += "\n"
-    ## Decays
+            namestr += " Q= " + _autostr(float(b.q), precision)
+        blockstr = "BLOCK %s\n" % namestr
+        entrystrs = []
+        for k, v in b.items():
+            entrystr = ""
+            if type(k) == tuple:
+                entrystr += sep.join(_autostr(i, precision) for i in k)
+            elif k is not None:
+                entrystr += _autostr(k, precision)
+            entrystr += sep + _autostr(v, precision)
+            entrystrs.append(entrystr)
+        blockstr += "\n".join(entrystrs)
+        blockstrs.append(blockstr)
+    return "\n\n".join(blockstrs)
+
+
+
+def writeSLHADecays(decays, ignorenobr=False, precision=8):
+    """Return an SLHA decay definition as a string, from the supplied decays dict."""
+    sep = 3 * " "
+    blockstrs = []
     for pid, particle in decays.iteritems():
-        out += ("DECAY %d " + fmte + "\n") % (particle.pid, particle.totalwidth or -1)
+        blockstr = ("DECAY %d " % particle.pid) + _autostr(particle.totalwidth or -1, precision) + "\n"
+        decaystrs = []
         for d in particle.decays:
             if d.br > 0.0 or not ignorenobr:
-                products_str = "   ".join(map(str, d.ids))
-                out += sep + fmte % d.br + sep + "%d" % len(d.ids) + sep + products_str + "\n"
-        out += "\n"
-    return out
+                products_str = sep.join(map(str, d.ids))
+                decaystr = sep + _autostr(d.br, precision) + sep + str(len(d.ids)) + sep + products_str
+                decaystrs.append(decaystr)
+        blockstr += "\n".join(decaystrs)
+        blockstrs.append(blockstr)
+    return "\n\n".join(blockstrs)
+
+
+
+def writeSLHA(blocks, decays, ignorenobr=False, precision=8):
+    """Return an SLHA definition as a string, from the supplied blocks and decays dicts."""
+    ss = [x for x in (writeSLHABlocks(blocks, precision), writeSLHADecays(decays, ignorenobr, precision)) if x]
+    return "\n\n".join(ss)
 
 
 
@@ -531,6 +648,7 @@ def readISAWIG(isastr, ignorenobr=False):
     def getnextvalidline():
         while LINES:
             s = LINES.pop(0).strip()
+            # print "*", s, "*"
             ## Return None if EOF reached
             if len(s) == 0:
                 continue
@@ -551,7 +669,7 @@ def readISAWIG(isastr, ignorenobr=False):
         hwid, mass, lifetime = getnextvalidlineitems()
         width = 1.0/(lifetime * 1.51926778e24) ## width in GeV == hbar/lifetime in seconds
         pdgid = herwigid2pdgid(hwid)
-        masses.add_entry((pdgid, mass))
+        masses[pdgid] = mass
         decays[pdgid] = Particle(pdgid, width, mass)
         #print pdgid, mass, width
     blocks["MASS"] = masses
@@ -580,59 +698,59 @@ def readISAWIG(isastr, ignorenobr=False):
     ## Now the SUSY parameters
     TANB, ALPHAH = getnextvalidlineitems()
     blocks["MINPAR"] = Block("MINPAR")
-    blocks["MINPAR"].add_entry((3, TANB))
+    blocks["MINPAR"][3] = TANB
     blocks["ALPHA"] = Block("ALPHA")
-    blocks["ALPHA"].entries = ALPHAH
+    blocks["ALPHA"][None] = ALPHAH
     #
     ## Neutralino mixing matrix
     blocks["NMIX"] = Block("NMIX")
     for i in xrange(1, 5):
         nmix_i = getnextvalidlineitems()
         for j, v in enumerate(nmix_i):
-            blocks["NMIX"].add_entry((i, j+1, v))
+            blocks["NMIX"][i, j+1] = v
     #
     ## Chargino mixing matrices V and U
     blocks["VMIX"] = Block("VMIX")
     vmix = getnextvalidlineitems()
-    blocks["VMIX"].add_entry((1, 1, vmix[0]))
-    blocks["VMIX"].add_entry((1, 2, vmix[1]))
-    blocks["VMIX"].add_entry((2, 1, vmix[2]))
-    blocks["VMIX"].add_entry((2, 2, vmix[3]))
+    blocks["VMIX"][1, 1] = vmix[0]
+    blocks["VMIX"][1, 2] = vmix[1]
+    blocks["VMIX"][2, 1] = vmix[2]
+    blocks["VMIX"][2, 2] = vmix[3]
     blocks["UMIX"] = Block("UMIX")
     umix = getnextvalidlineitems()
-    blocks["UMIX"].add_entry((1, 1, umix[0]))
-    blocks["UMIX"].add_entry((1, 2, umix[1]))
-    blocks["UMIX"].add_entry((2, 1, umix[2]))
-    blocks["UMIX"].add_entry((2, 2, umix[3]))
+    blocks["UMIX"][1, 1] = umix[0]
+    blocks["UMIX"][1, 2] = umix[1]
+    blocks["UMIX"][2, 1] = umix[2]
+    blocks["UMIX"][2, 2] = umix[3]
     #
     THETAT, THETAB, THETAL = getnextvalidlineitems()
     import math
     blocks["STOPMIX"] = Block("STOPMIX")
-    blocks["STOPMIX"].add_entry((1, 1,  math.cos(THETAT)))
-    blocks["STOPMIX"].add_entry((1, 2, -math.sin(THETAT)))
-    blocks["STOPMIX"].add_entry((2, 1,  math.sin(THETAT)))
-    blocks["STOPMIX"].add_entry((2, 2,  math.cos(THETAT)))
+    blocks["STOPMIX"][1, 1] =  math.cos(THETAT)
+    blocks["STOPMIX"][1, 2] = -math.sin(THETAT)
+    blocks["STOPMIX"][2, 1] =  math.sin(THETAT)
+    blocks["STOPMIX"][2, 2] =  math.cos(THETAT)
     blocks["SBOTMIX"] = Block("SBOTMIX")
-    blocks["SBOTMIX"].add_entry((1, 1,  math.cos(THETAB)))
-    blocks["SBOTMIX"].add_entry((1, 2, -math.sin(THETAB)))
-    blocks["SBOTMIX"].add_entry((2, 1,  math.sin(THETAB)))
-    blocks["SBOTMIX"].add_entry((2, 2,  math.cos(THETAB)))
+    blocks["SBOTMIX"][1, 1] =  math.cos(THETAB)
+    blocks["SBOTMIX"][1, 2] = -math.sin(THETAB)
+    blocks["SBOTMIX"][2, 1] =  math.sin(THETAB)
+    blocks["SBOTMIX"][2, 2] =  math.cos(THETAB)
     blocks["STAUMIX"] = Block("STAUMIX")
-    blocks["STAUMIX"].add_entry((1, 1,  math.cos(THETAL)))
-    blocks["STAUMIX"].add_entry((1, 2, -math.sin(THETAL)))
-    blocks["STAUMIX"].add_entry((2, 1,  math.sin(THETAL)))
-    blocks["STAUMIX"].add_entry((2, 2,  math.cos(THETAL)))
+    blocks["STAUMIX"][1, 1] =  math.cos(THETAL)
+    blocks["STAUMIX"][1, 2] = -math.sin(THETAL)
+    blocks["STAUMIX"][2, 1] =  math.sin(THETAL)
+    blocks["STAUMIX"][2, 2] =  math.cos(THETAL)
     #
     ATSS, ABSS, ALSS = getnextvalidlineitems()
     blocks["AU"] = Block("AU")
-    blocks["AU"].add_entry((3, 3, ATSS))
+    blocks["AU"][3, 3] = ATSS
     blocks["AD"] = Block("AD")
-    blocks["AD"].add_entry((3, 3, ABSS))
+    blocks["AD"][3, 3] = ABSS
     blocks["AE"] = Block("AE")
-    blocks["AE"].add_entry((3, 3, ALSS))
+    blocks["AE"][3, 3] = ALSS
     #
     MUSS = getnextvalidlineitems()[0]
-    blocks["MINPAR"].add_entry((4, MUSS))
+    blocks["MINPAR"][4] = MUSS
     #
 
     # TODO: Parse RPV boolean and couplings into SLHA2 blocks
@@ -640,58 +758,6 @@ def readISAWIG(isastr, ignorenobr=False):
     return blocks, decays
 
 
-def writeISAJET(blocks, decays, outname, ignorenobr=False, precision=8):
-    """
-    Return a SUSY spectrum definition in the format required for input by ISAJET,
-    as a string, from the supplied blocks and decays dicts.
-
-    The outname parameter specifies the desired output filename from ISAJET: this
-    will appear in the first line of the return value.
-
-    If the ignorenobr parameter is True, do not write decay entries with a
-    branching ratio of zero.
-    """
-    fmte = "%." + str(precision) + "e"
-
-    masses = blocks["MASS"].entries
-
-    ## Init output string
-    out = ""
-
-    ## First line is the output name
-    out += "'%s'" % outname + "\n"
-
-    ## Next the top mass
-    out += fmte % masses[6] + "\n"
-
-    ## Next the top mass
-    out += fmte % masses[6] + "\n"
-
-    ## mSUGRA parameters (one line)
-    # e.g. 1273.78,713.286,804.721,4.82337
-
-    ## Masses and trilinear couplings (3 lines)
-    # e.g. 1163.14,1114.15,1118.99,374.664,209.593
-    # e.g. 1069.54,1112.7,919.908,374.556,209.381,-972.817,-326.745,-406.494
-    # e.g. 1163.14,1114.15,1118.99,374.712,210.328
-
-    ## RPV couplings (?? lines)
-    # e.g. 232.615,445.477
-
-    ## Etc ???!!!
-    # e.g. /
-    # e.g. n
-    # e.g. y
-    # e.g. y
-    # e.g. 0.047441 3.80202e-23 0 0 0 2.17356e-22 0 0 5.23773e-09
-    # e.g. y
-    # e.g. 3.35297e-25 0 0 0 7.34125e-24 0 0 0 3.17951e-22 8.07984e-12 0 0 0 1.76906e-10 0 0 0 7.66184e-09 0 0 0 0 0 0 0 0 0
-    # e.g. n
-    # e.g. 'susy_RPV_stau_BC1scan_m560_tanb05.txt'
-
-    return out
-
-
 def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
     """
     Return a SUSY spectrum definition in the format produced by ISAWIG for inut to HERWIG
@@ -703,9 +769,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
     If the ignorenobr parameter is True, do not write decay entries with a
     branching ratio of zero.
     """
-    fmte = "%." + str(precision) + "e"
-
-    masses = blocks["MASS"].entries
+    masses = blocks["MASS"]
 
     ## Init output string
     out = ""
@@ -725,7 +789,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
                 lifetime = 1.0/(width * 1.51926778e24) ## lifetime in seconds == hbar/width in GeV
         except:
             pass
-        massout += ("%d " + fmte + " " + fmte + "\n") % (pdgid2herwigid(pid), masses[pid], lifetime)
+        massout += ("%d " % pdgid2herwigid(pid)) + _autostr(masses[pid], precision) + " " + _autostr(lifetime, precision) + "\n"
     out += "%d\n" % massout.count("\n")
     out += massout
 
@@ -763,7 +827,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
             if len(d.ids) == 3:
                 # TODO: How to determine the conditions for using 200 and 300 MEs? Enumeration of affected decays?
                 pass
-            decayout += "%d " + fmte + " %d " % (hwid, d.br, nme)
+            decayout += ("%d " % hwid) + _autostr(d.br, precision) + (" %d " % nme)
 
             def is_quark(pid):
                 return (abs(pid) in range(1, 7))
@@ -810,11 +874,11 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
                 if len(absids) == 2:
                     ## 2 body mode, to Higgs: Higgs; Bottom
                     if (25 in absids or 26 in absids) and 5 in absids:
-                        d.ids = sorted(d.ids, key=cmp_bottomlast)
+                        d.ids = sorted(d.ids, cmp=cmp_bottomlast)
                 elif len(absids) == 3:
                     ## 3 body mode, via charged Higgs/W: quarks or leptons from W/Higgs; Bottom
                     if 37 in absids or 23 in absids:
-                        d.ids = sorted(d.ids, key=cmp_bottomlast)
+                        d.ids = sorted(d.ids, cmp=cmp_bottomlast)
             ## Gluino
             elif abs(pid) == 1000021:
                 if len(absids) == 2:
@@ -829,7 +893,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
                             if b == 21:
                                 return True
                             return cmp(a, b)
-                        d.ids = sorted(d.ids, key=cmp_gluonfirst)
+                        d.ids = sorted(d.ids, cmp=cmp_gluonfirst)
                 elif len(absids) == 3:
                     ## 3-body modes, R-parity conserved: colour neutral; q or qbar
                     def cmp_quarkslast(a, b):
@@ -839,7 +903,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
                         if is_quark(b):
                             return False
                         return cmp(a, b)
-                    d.ids = sorted(d.ids, key=cmp_quarkslast)
+                    d.ids = sorted(d.ids, cmp=cmp_quarkslast)
             ## Squark/Slepton
             elif is_squark(pid) or is_slepton(pid):
                 def cmp_susy_quark_lepton(a, b):
@@ -860,7 +924,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
                 ##                 Baryon number violated               quark     quark
                 ## Slepton
                 ##   2 body modes: Lepton Number Violated               q or qbar
-                d.ids = sorted(d.ids, key=cmp_bottomlast)
+                d.ids = sorted(d.ids, cmp=cmp_bottomlast)
             ## Higgs
             elif pid in (25, 26):
                 # TODO: Includes SUSY Higgses?
@@ -876,7 +940,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
                         if is_quark(b):
                             return False
                         return cmp(a, b)
-                    d.ids = sorted(d.ids, key=cmp_quarkslast)
+                    d.ids = sorted(d.ids, cmp=cmp_quarkslast)
             elif is_gaugino(pid):
                 # TODO: Is there actually anything to do here?
                 ## Gaugino
@@ -893,7 +957,7 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
                         if is_quark(b):
                             return False
                         return cmp(a, b)
-                    d.ids = sorted(d.ids, key=cmp_quarkslast)
+                    d.ids = sorted(d.ids, cmp=cmp_quarkslast)
 
             # TODO: Gaugino/Gluino
             ##   3 body modes:  R-parity violating:   Particles in the same order as the R-parity violating superpotential
@@ -909,27 +973,36 @@ def writeISAWIG(blocks, decays, ignorenobr=False, precision=8):
 
     ## Now the SUSY parameters
     ## TANB, ALPHAH:
-    out += (fmte + " " + fmte + "\n") % (blocks["MINPAR"].entries[3], blocks["ALPHA"].entries)
+    out += _autostr(blocks["MINPAR"][3], precision) + " " + _autostr(blocks["ALPHA"].value(), precision) + "\n"
     ## Neutralino mixing matrix
-    nmix = blocks["NMIX"].entries
+    nmix = blocks["NMIX"]
     for i in xrange(1, 5):
-        out += (fmte + " " + fmte + " " + fmte + " " + fmte + "\n") % (nmix[i][1], nmix[i][2], nmix[i][3], nmix[i][4])
+        out += _autostr(nmix[i,1], precision) + " " + \
+               _autostr(nmix[i,2], precision) + " " + \
+               _autostr(nmix[i,3], precision) + " " + \
+               _autostr(nmix[i,4], precision) + "\n"
     ## Chargino mixing matrices V and U
-    vmix = blocks["VMIX"].entries
-    out += (fmte + " " + fmte + " " + fmte + " " + fmte + "\n") % (vmix[1][1], vmix[1][2], vmix[2][1], vmix[2][2])
-    umix = blocks["UMIX"].entries
-    out += (fmte + " " + fmte + " " + fmte + " " + fmte + "\n") % (umix[1][1], umix[1][2], umix[2][1], umix[2][2])
+    vmix = blocks["VMIX"]
+    out += _autostr(vmix[1,1], precision) + " " + \
+           _autostr(vmix[1,2], precision) + " " + \
+           _autostr(vmix[2,1], precision) + " " + \
+           _autostr(vmix[2,2], precision) + "\n"
+    umix = blocks["UMIX"]
+    out += _autostr(umix[1,1], precision) + " " + \
+           _autostr(umix[1,2], precision) + " " + \
+           _autostr(umix[2,1], precision) + " " + \
+           _autostr(umix[2,2], precision) + "\n"
     ## THETAT,THETAB,THETAL
     import math
-    out += (fmte + " " + fmte + " " + fmte + " " + "\n") % (math.acos(blocks["STOPMIX"].entries[1][1]),
-                                                            math.acos(blocks["SBOTMIX"].entries[1][1]),
-                                                            math.acos(blocks["STAUMIX"].entries[1][1]))
+    out += _autostr(math.acos(blocks["STOPMIX"][1,1]), precision) + " " + \
+           _autostr(math.acos(blocks["SBOTMIX"][1,1]), precision) + " " + \
+           _autostr(math.acos(blocks["STAUMIX"][1,1]), precision) + "\n"
     ## ATSS,ABSS,ALSS
-    out += (fmte + " " + fmte + " " + fmte + " " + "\n") % (blocks["AU"].entries[3][3],
-                                                            blocks["AD"].entries[3][3],
-                                                            blocks["AE"].entries[3][3])
+    out += _autostr(blocks["AU"][3,3], precision) + " " + \
+           _autostr(blocks["AD"][3,3], precision) + " " + \
+           _autostr(blocks["AE"][3,3], precision) + "\n"
     ## MUSS == sign(mu)
-    out += "%f\n" % blocks["MINPAR"].entries[4]
+    out += "%f\n" % blocks["MINPAR"][4]
 
     ## RPV SUSY
     isRPV = False
@@ -1024,7 +1097,7 @@ if __name__ == "__main__":
 
         print blocks.keys()
 
-        print blocks["MASS"].entries[25]
+        print blocks["MASS"][25]
         print
 
         for p in sorted(decays.values()):
